# tenant_context.py

from contextlib import contextmanager
from contextvars import ContextVar, Token

from fastapi import HTTPException, Request

from .service import TenantConfigService


_current_request_id: ContextVar[str | None] = ContextVar("request_id", default=None)
_current_tenant: ContextVar[str | None] = ContextVar("tenant", default=None)
_current_user_id: ContextVar[str | None] = ContextVar("user_id", default=None)


class TenantContextConfig:
    def __init__(
        self,
        tenants_source: str,
        *,
        tenant_header: str = "X-Tenant",
        user_id_header: str = "X-UserId",
        request_id_header: str = "X-Request-Id",
        default_tenant: str | None = None,
        default_user_id: str | None = None,
        allowed_tenants: set[str] | None = None,
        strict_whitelist: bool = False,
    ):
        self.tenant_header = tenant_header
        self.user_id_header = user_id_header
        self.request_id_header = request_id_header

        self.default_tenant = default_tenant
        self.default_user_id = default_user_id

        self.allowed_tenants = allowed_tenants
        self.strict_whitelist = strict_whitelist

        self.service = TenantConfigService(tenants_source)

    # -------------------------
    # TENANT
    # -------------------------

    def get_tenant(self, request: Request) -> str:
        tenant = request.headers.get(self.tenant_header)

        if tenant and tenant.strip():
            tenant = tenant.strip().lower()
        else:
            tenant = self.default_tenant

        if not tenant:
            raise HTTPException(400, f"Header {self.tenant_header} mancante")

        if self.allowed_tenants and self.strict_whitelist:
            if tenant not in self.allowed_tenants:
                raise HTTPException(403, "Tenant non consentito")

        try:
            self.service.get(tenant)
        except FileNotFoundError:
            raise HTTPException(404, f"Tenant '{tenant}' non trovato")

        return tenant

    def get_tenant_config(self, request: Request):
        return self.service.get(self.get_tenant(request))

    # -------------------------
    # HEADERS
    # -------------------------

    def get_user_id(self, request: Request) -> str | None:
        user_id = request.headers.get(self.user_id_header)
        return user_id.strip() if user_id and user_id.strip() else self.default_user_id

    def get_request_id(self, request: Request) -> str | None:
        request_id = request.headers.get(self.request_id_header)
        return request_id.strip() if request_id and request_id.strip() else None

    # -------------------------
    # CONTEXT
    # -------------------------

    @contextmanager
    def tenant_context(self, tenant: str | None):
        token: Token = _current_tenant.set(tenant)
        try:
            yield
        finally:
            _current_tenant.reset(token)

    @contextmanager
    def user_id_context(self, user_id: str | None):
        token: Token = _current_user_id.set(user_id)
        try:
            yield
        finally:
            _current_user_id.reset(token)

    @contextmanager
    def request_id_context(self, request_id: str | None):
        token: Token = _current_request_id.set(request_id)
        try:
            yield
        finally:
            _current_request_id.reset(token)

    # -------------------------
    # CURRENT
    # -------------------------

    def current_tenant(self) -> str | None:
        return _current_tenant.get()

    def current_user_id(self) -> str | None:
        return _current_user_id.get()

    def current_request_id(self) -> str | None:
        return _current_request_id.get()


# tenant_db.py

from contextlib import asynccontextmanager
from contextvars import ContextVar, Token
from typing import AsyncIterator

from fastapi import HTTPException, Request
from sqlalchemy.ext.asyncio import AsyncSession

from .db_engine import TenantDatabaseManager
from .tenant_context import TenantContextConfig
from .exceptions import TenantDatabaseConfigError, TenantDatabaseError


_current_db: ContextVar[AsyncSession | None] = ContextVar(
    "current_db",
    default=None,
)


class TenantDbConfig(TenantContextConfig):
    def __init__(
        self,
        tenants_source: str,
        *,
        tenant_header: str = "X-Tenant",
        user_id_header: str = "X-UserId",
        request_id_header: str = "X-Request-Id",
        default_tenant: str | None = None,
        default_user_id: str | None = None,
        allowed_tenants: set[str] | None = None,
        strict_whitelist: bool = False,
        mode: str = "testing",
        echo_sql: bool = False,
        pool_size: int = 1,
        max_overflow: int = 2,
        pool_timeout: int = 5,
        pool_pre_ping: bool = True,
        app_name_prefix: str = "fastapi",
    ):
        super().__init__(
            tenants_source=tenants_source,
            tenant_header=tenant_header,
            user_id_header=user_id_header,
            request_id_header=request_id_header,
            default_tenant=default_tenant,
            default_user_id=default_user_id,
            allowed_tenants=allowed_tenants,
            strict_whitelist=strict_whitelist,
        )

        self.mode = mode
        self.echo_sql = echo_sql
        self.pool_size = pool_size
        self.max_overflow = max_overflow
        self.pool_timeout = pool_timeout
        self.pool_pre_ping = pool_pre_ping
        self.app_name_prefix = app_name_prefix

        self.db = TenantDatabaseManager(self)

    async def open_db(self, request: Request) -> AsyncIterator[None]:
        tenant = self.get_tenant(request)
        user_id = self.get_user_id(request)
        request_id = self.get_request_id(request)
        try:
            sessionmaker = await self.db.get_sessionmaker(tenant)
        except TenantDatabaseError as exc:
            raise HTTPException(
                status_code=503,
                detail=f"Database non disponibile per il tenant '{tenant}'",
            ) from exc
            
        with (
            self.tenant_context(tenant),
            self.user_id_context(user_id),
            self.request_id_context(request_id),
        ):
            async with sessionmaker() as db:
                token: Token = _current_db.set(db)
                try:
                    yield
                except Exception:
                    if db.in_transaction():
                        await db.rollback()
                    raise
                finally:
                    _current_db.reset(token)

    def current_db(self) -> AsyncSession:
        db = _current_db.get()
        if db is None:
            raise RuntimeError("Nessun DB attivo nel contesto corrente")
        return db

    async def commit(self) -> None:
        await self.current_db().commit()

    async def rollback(self) -> None:
        db = self.current_db()
        if db.in_transaction():
            await db.rollback()

    async def flush(self) -> None:
        await self.current_db().flush()

    @asynccontextmanager
    async def transaction(self):
        try:
            yield self.current_db()
            await self.commit()
        except Exception:
            await self.rollback()
            raise

    async def dispose(self) -> None:
        await self.db.dispose_all()


# fastapi.py
from fastapi.responses import JSONResponse
from sqlalchemy.exc import DBAPIError, OperationalError, StatementError


async def database_error_middleware(request, call_next):
    try:
        return await call_next(request)

    except OperationalError:
        return JSONResponse(
            status_code=503,
            content={"detail": "Database non disponibile"},
        )

    except StatementError:
        return JSONResponse(
            status_code=503,
            content={"detail": "Errore query"},
        )

    except DBAPIError:
        return JSONResponse(
            status_code=503,
            content={"detail": "Errore database"},
        )


from fastapi import APIRouter, Depends, HTTPException
from typing import Type, TypeAlias, Callable, Any
from uuid import UUID
from sqlalchemy.ext.asyncio import AsyncSession
from .response_schema import IDeleteResponseBase, IGetResponseBase, IPostResponseBase, IPutResponseBase, IResponseList, create_list_response, create_response
### idea da sviluppare......

def create_crud_router(
    *,
    crud,
    prefix: str = "",
    tags: list[str] | None = None,
    open_db_dependency: Callable = None,
    enable_list: bool = True,
    enable_get: bool = True,
    enable_create: bool = True,
    enable_update: bool = True,
    enable_delete: bool = True,
    enable_soft_delete: bool = False,
):
    """
    Factory per generare router CRUD standardizzati.

    Pensata per:
    - CRUDBase
    - IResponseBase / IGetResponseBase / IResponseList
    - TenantDbConfig.open_db
    """

    router = APIRouter(prefix=prefix, tags=tags or [])

    # fallback dependency (se non passato)
    db_dep = open_db_dependency or (lambda: None)

    # -------------------------
    # LIST
    # -------------------------
    if enable_list:
        @router.get("/", response_model=IResponseList[read_schema])
        async def list_items(_: None = Depends(db_dep)):
            items = await crud.get_multi()
            return create_list_response(items)

    # -------------------------
    # GET
    # -------------------------
    if enable_get:
        @router.get("/{id}", response_model=IGetResponseBase[read_schema])
        async def get_item(
            id: UUID | str | int,
            _: None = Depends(db_dep),
        ):
            obj = await crud.get(id=id)

            if not obj:
                raise HTTPException(status_code=404, detail="Not found")

            return create_response(obj, response_model=IGetResponseBase)

    # -------------------------
    # CREATE
    # -------------------------
    if enable_create:
        @router.post("/", response_model=IPostResponseBase[read_schema])
        async def create_item(
            payload: create_schema,
            db: AsyncSession = Depends(db_dep),
        ):
            obj = await crud.create(obj_in=payload, db=db)

            return create_response(obj, response_model=IPostResponseBase)

    # -------------------------
    # UPDATE
    # -------------------------
    if enable_update:
        @router.put("/{id}", response_model=IGetResponseBase[read_schema])
        async def update_item(
            id: UUID | str | int,
            payload: update_schema,
            db: AsyncSession = Depends(db_dep),
        ):
            obj = await crud.get_required(id=id, db=db)

            obj = await crud.update(
                db_obj=obj,
                obj_in=payload,
                db=db,
            )

            return create_response(obj, response_model=IPutResponseBase)

    # -------------------------
    # DELETE
    # -------------------------
    if enable_delete:
        @router.delete("/{id}", response_model=IDeleteResponseBase[None])
        async def delete_item(
            id: UUID | str | int,
            db: AsyncSession = Depends(db_dep),
        ):
            deleted = await crud.delete(id=id, db=db)

            if not deleted:
                raise HTTPException(status_code=404, detail="Not found")

            return create_response(None, response_model=IDeleteResponseBase)

    # -------------------------
    # SOFT DELETE (opzionale)
    # -------------------------
    if enable_soft_delete:
        @router.delete("/soft/{id}", response_model=IDeleteResponseBase[None])
        async def soft_delete_item(
            id: UUID | str | int,
            db: AsyncSession = Depends(db_dep),
        ):
            deleted = await crud.soft_delete(id=id, db=db)

            if not deleted:
                raise HTTPException(status_code=404, detail="Not found")

            return create_response(None, response_model=IDeleteResponseBase)

    return router


#app/crud/base_crud.py
from datetime import datetime
from enum import Enum
from fastapi import HTTPException, Query
from typing import Any, Dict, Generic, Optional, Sequence, Type, TypeVar
from uuid import UUID

from psycopg import IntegrityError
from fastapi_pagination.ext.sqlalchemy import paginate
from fastapi_pagination import Params, Page
from pydantic import BaseModel
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy import Select, exc, func, inspect, select
from fastapi_pagination.bases import AbstractPage, AbstractParams
from sqlalchemy.orm import DeclarativeMeta

ModelType = TypeVar("ModelType", bound=DeclarativeMeta)
CreateSchemaType = TypeVar("CreateSchemaType", bound=BaseModel)
UpdateSchemaType = TypeVar("UpdateSchemaType", bound=BaseModel)
SchemaType = TypeVar("SchemaType", bound=DeclarativeMeta)
T = TypeVar("T", bound=DeclarativeMeta)


class IOrderEnum(str, Enum):
    ascendent = "ascendent"
    descendent = "descendent"


class MyPage(AbstractPage[T], Generic[T]):
    results: list[T]
    totalResults: int

    __params_type__ = Params

    @classmethod
    def create(
        cls,
        items: Sequence[T],
        params: AbstractParams,
        *,
        total: Optional[int] = None,
        **kwargs: Any
    ) -> Page[T]:
        assert total is not None, "total must be provided"

        return cls(
            results=items,
            totalResults=total,
        )


class CRUDBase(Generic[ModelType, CreateSchemaType, UpdateSchemaType]):
    def __init__(self, model: type[ModelType]):
        self.model = model
    
    async def get(
        self, *, id: UUID | str, db_session: AsyncSession
    ) -> ModelType | None:

        query = select(self.model).where(self.model.id == id)
        response = await db_session.execute(query)
        return response.scalar_one_or_none()

    async def get_by_ids(
        self,
        *,
        list_ids: list[UUID | str],
        db_session: AsyncSession,
    ) -> list[ModelType] | None:

        response = await db_session.execute(
            select(self.model).where(self.model.id.in_(list_ids))
        )
        return response.scalars().all()

    async def get_count(
        self, *, db_session: AsyncSession
    ) -> ModelType | None:

        response = await db_session.execute(
            select(func.count()).select_from(select(self.model).subquery())
        )
        return response.scalar_one()

    async def get_multi(
        self,
        *,
        skip: int = 0,
        limit: int = 100,
        query: T | Select[T] | None = None,
        db_session: AsyncSession,
    ) -> list[ModelType]:
                        
        if query is None:
            query = select(self.model).offset(skip).limit(limit).order_by(self.model.id)
        response = await db_session.execute(query)
        return response.all()

    async def get_multi_all(
        self,
        *,
        query: T | Select[T] | None = None,
        db_session: AsyncSession,
    ) -> list[ModelType]:

        response = await db_session.execute(query)
        return response.all()
    
    async def get_multi_mappings(
        self,
        *,
        query: T | Select[T] | None = None,
        db_session: AsyncSession,
    ) -> list[ModelType]:
        db_session = db_session or self.db.session
        response = await db_session.execute(query)
        return response.mappings().all()

    async def get_multi_paginated(
        self,
        *,
        params: Params | None = Params(),
        query: T | Select[T] | None = None,
        db_session: AsyncSession,
    ) -> Page[ModelType]:
        
        if query is None:
            query = select(self.model)
        result = await paginate(db_session, query, params)
        return result
    

    async def get_multi_paginated_ordered(
        self,
        *,
        params: Params | None = Params(),
        sort: str | None = None,
        db_session: AsyncSession,
        order: IOrderEnum | None = IOrderEnum.ascendent,
        query: T | Select[T] | None = None,
    ) -> Page[ModelType]:
        
        columns = self.model.__table__.columns

        if sort is None or sort not in columns:
            sort = "id"

        if query is None:
            if order == IOrderEnum.ascendent:
                query = select(self.model).order_by(columns[sort].asc())
            else:
                query = select(self.model).order_by(columns[sort].desc())

        return await paginate(db_session, query, params)

    async def get_multi_ordered(
        self,
        *,
        skip: int = 0,
        limit: int = 100,
        sort: str | None = None,
        order: IOrderEnum | None = IOrderEnum.ascendent,
        query: T | Select[T] | None = None,
        db_session: AsyncSession,
    ) -> list[ModelType]:

        columns = self.model.__table__.columns

        if sort is None or sort not in columns:
            sort = "id"

        if query is None:
            if order == IOrderEnum.ascendent:
                query = select(self.model).offset(skip).limit(limit).order_by(columns[sort].asc())
            else:
                query = select(self.model).offset(skip).limit(limit).order_by(columns[sort].desc())

        response = await db_session.execute(query)
        return response.scalars().all()

    async def create(
        self,
        *,
        obj_in: CreateSchemaType | ModelType,
        created_by_id: UUID | str | None = None,
        db_session: AsyncSession,
    ) -> ModelType:

        if isinstance(obj_in, self.model):
            db_obj = obj_in
        else:
            # compatibile con Pydantic v2
            db_obj = self.model(**obj_in.model_dump())

        if created_by_id and hasattr(db_obj, "created_by_id"):
            setattr(db_obj, "created_by_id", created_by_id)

        try:
            db_session.add(db_obj)
            await db_session.commit()
            await db_session.refresh(db_obj)
        except IntegrityError as e:
            await db_session.rollback()
            raise HTTPException(
                status_code=409,
                detail="Resource already exists",
            ) from e

        return db_obj

    async def update(
        self,
        *,
        obj_current: ModelType,
        obj_in: UpdateSchemaType | dict[str, Any] | ModelType,
        updated_by_id: UUID | str | None = None,
        db_session: AsyncSession,
    ) -> ModelType:        


        if isinstance(obj_in, dict):
            update_data = obj_in
        else:
            update_data = obj_in.model_dump(
                exclude_unset=True
            )
            

        #for field in update_data:
        #    setattr(obj_current, field, update_data[field])
        
        setattr(obj_current, 'jd', update_data)
        if updated_by_id:
            setattr(obj_current, 'updated_by_id', updated_by_id)
            setattr(obj_current, 'updated_updated_atby_id', datetime.now())
        
        db_session.add(obj_current)
        await db_session.commit()
        await db_session.refresh(obj_current)
        return obj_current
    
 
    async def update_model_recursive(
        self,
        *,
        model: Type[T],
        existing_instance: Optional[T],
        obj_in: UpdateSchemaType | dict[str, Any] | ModelType
    ) -> T:
        if isinstance(obj_in, dict):
            data = obj_in
        else:
            data = obj_in.model_dump(exclude_unset=True)

        relationship_fields = {rel.key for rel in model.__mapper__.relationships}
        direct_fields = {k: v for k, v in data.items() if k not in relationship_fields}

        instance = existing_instance or model(**direct_fields)
        if existing_instance:
            for k, v in direct_fields.items():
                setattr(instance, k, v)

        mapper = inspect(model)

        for field_name, value in data.items():
            if field_name not in direct_fields:
                rel = mapper.relationships.get(field_name)
                if not rel:
                    continue
                related_model = rel.entity.class_

                if rel.uselist:  # uno-a-molti
                    existing_items = {getattr(i, "id", None): i for i in getattr(instance, field_name, [])}
                    new_items = []
                    for item in value:
                        item_id = item.get("id")
                        if item_id and item_id in existing_items:
                            updated = await self.update_model_recursive(
                                model=related_model,
                                existing_instance=existing_items[item_id],
                                obj_in=item
                            )
                        else:
                            updated = related_model(**{k: v for k, v in item.items() if not isinstance(v, (dict, list))})
                            for fk in related_model.__table__.foreign_keys:
                                if fk.column.table == model.__table__:
                                    setattr(updated, fk.parent.name, instance.id)
                        new_items.append(updated)
                    setattr(instance, field_name, new_items)

                else:  # uno-a-uno (es. dichiarante, procuratore)
                    sub_instance = getattr(instance, field_name, None)
                    if value:
                        updated_sub = await self.update_model_recursive(
                            model=related_model,
                            existing_instance=sub_instance,
                            obj_in=value
                        )
                        # collega subito la chiave esterna
                        for fk in related_model.__table__.foreign_keys:
                            if fk.column.table == model.__table__:
                                setattr(updated_sub, fk.parent.name, instance.id)
                        setattr(instance, field_name, updated_sub)

        return instance
    

    
    async def update_all(
        self,
        *,
        obj_current: ModelType,
        obj_in: UpdateSchemaType | dict[str, Any] | ModelType,
        updated_by_id: UUID | str | None = None,
        db_session: AsyncSession,
    ) -> ModelType:
                
        if updated_by_id:
            setattr(obj_in,"updated_by_id",updated_by_id) 
            setattr(obj_in,"updated_at",datetime.now()) 
        
        model = type(obj_current)
        istanza = await self.update_model_recursive(model=model,existing_instance=obj_current,obj_in=obj_in)    
        
        db_session.add(istanza)
        await db_session.commit()
        await db_session.refresh(istanza)
        return obj_current
    

    async def remove(
        self, *, id: UUID | str, db_session: AsyncSession
    ) -> ModelType:

        response = await db_session.execute(
            select(self.model).where(self.model.id == id)
        )
        obj = response.scalar_one()
        await db_session.delete(obj)
        await db_session.commit()
        return obj

    async def remove_soft(self, *, id: UUID | str, db_session: AsyncSession) -> ModelType:
        obj = await self.get(id=id, db_session=db_session)
        obj.is_active = False
        obj.deleted_at = datetime.now()
        db_session.add(obj)
        await db_session.commit()
        return obj



# tenant_oidc.py

# Questa classe:

# estende TenantContextConfig
# aggiunge TenantRuntimeConfigService
# espone getter runtime per tenant:
# get_oidc()
# get_email()
# get_pagopa()
# get_protocollo()


from .models import KeycloakAdminConfig
from .tenant_context import TenantContextConfig
from .service import TenantRuntimeConfigService


class TenantRuntimeConfig(TenantContextConfig):
    def __init__(
        self,
        tenants_source: str,
        *,
        settings=None,
        tenant_header: str = "X-Tenant",
        user_id_header: str = "X-UserId",
        request_id_header: str = "X-Request-Id",
        default_tenant: str | None = None,
        default_user_id: str | None = None,
        allowed_tenants: set[str] | None = None,
        strict_whitelist: bool = False,
    ):
        super().__init__(
            tenants_source=tenants_source,
            tenant_header=tenant_header,
            user_id_header=user_id_header,
            request_id_header=request_id_header,
            default_tenant=default_tenant,
            default_user_id=default_user_id,
            allowed_tenants=allowed_tenants,
            strict_whitelist=strict_whitelist,
        )

        self.runtime = TenantRuntimeConfigService(self.service, settings)

    # -------------------------
    # RUNTIME CONFIG
    # -------------------------

    def get_oidc(self, tenant: str):
        return self.runtime.get_oidc(tenant)
    
    def get_keycloak_admin(self, tenant: str):
        return self.runtime.get_keycloak_admin(tenant)

    def get_email(self, tenant: str):
        return self.runtime.get_email(tenant)

    def get_pagopa(self, tenant: str):
        return self.runtime.get_pagopa(tenant)

    def get_protocollo(self, tenant: str):
        return self.runtime.get_protocollo(tenant)


from math import ceil
from typing import Any, Generic, TypeVar
from collections.abc import Sequence

from fastapi import Query, Response
from fastapi_pagination import Page
from fastapi_pagination.bases import AbstractPage, AbstractParams, RawParams
from pydantic import BaseModel, Field, ConfigDict

DataType = TypeVar("DataType")
T = TypeVar("T")


class XMLResponse(Response):
    media_type = "application/xml"


# =========================
# PAGINATION PARAMS
# =========================
class PageParams(BaseModel, AbstractParams):
    page: int = Query(1, ge=1)
    size: int = Query(10, ge=1, le=10000)

    def to_raw_params(self) -> RawParams:
        return RawParams(
            limit=self.size,
            offset=(self.page - 1) * self.size,
        )


# =========================
# PAGE BASE
# =========================
class PageBase(Page[T], Generic[T]):
    previous_page: int | None = Field(
        default=None,
        description="Page number of the previous page",
    )
    next_page: int | None = Field(
        default=None,
        description="Page number of the next page",
    )


# =========================
# BASE RESPONSES
# =========================
class IResponseList(BaseModel, Generic[T]):
    model_config = ConfigDict(from_attributes=True)

    message: str = ""
    meta: dict[str, Any] = Field(default_factory=dict)
    data: list[T] | None = None


class IResponseBase(BaseModel, Generic[T]):
    model_config = ConfigDict(from_attributes=True)

    message: str = ""
    meta: dict[str, Any] = Field(default_factory=dict)
    data: T | None = None


# =========================
# PAGINATED RESPONSE
# =========================
class IGetResponsePaginated(AbstractPage[T], Generic[T]):
    model_config = ConfigDict(from_attributes=True)

    message: str = ""
    meta: dict[str, Any] = Field(default_factory=dict)
    data: PageBase[T]

    __params_type__ = PageParams  # 👈 fondamentale

    @classmethod
    def create(
        cls,
        items: Sequence[T],
        total: int,
        params: AbstractParams,
    ) -> "IGetResponsePaginated[T]":
        raw_params = params.to_raw_params()

        size = raw_params.limit or 10
        offset = raw_params.offset or 0

        page = (offset // size) + 1 if size > 0 else 1
        pages = ceil(total / size) if size > 0 else 0

        return cls(
            message="Data paginated correctly",
            meta={},
            data=PageBase[T](
                items=list(items),
                page=page,
                size=size,
                total=total,
                pages=pages,
                next_page=page + 1 if page < pages else None,
                previous_page=page - 1 if page > 1 else None,
            ),
        )


# =========================
# SPECIALIZED RESPONSES
# =========================
class IGetResponseBase(IResponseBase[DataType], Generic[DataType]):
    message: str = "Data got correctly"


class IPostResponseBase(IResponseBase[DataType], Generic[DataType]):
    message: str = "Data created correctly"


class IPutResponseBase(IResponseBase[DataType], Generic[DataType]):
    message: str = "Data updated correctly"


class IPatchResponseBase(IResponseBase[DataType], Generic[DataType]):
    message: str = "Data updated correctly"


class IDeleteResponseBase(IResponseBase[DataType], Generic[DataType]):
    message: str = "Data deleted correctly"


# =========================
# RESPONSE FACTORY
# =========================
def create_response(
    data: DataType,
    message: str | None = None,
    meta: dict | Any | None = {},
) -> (
    IResponseBase[DataType]
    | IGetResponsePaginated[DataType]
    | IGetResponseBase[DataType]
    | IPutResponseBase[DataType]
    | IDeleteResponseBase[DataType]
    | IPostResponseBase[DataType]
):
    if isinstance(data, IGetResponsePaginated):
        data.message = "Data paginated correctly" if message is None else message
        data.meta = meta
        return data
    if message is None:
        return {"data": data, "meta": meta}
    return {"data": data, "message": message, "meta": meta}


def create_list_response(
    data: list[DataType],
    message: str | None = None,
    meta: dict[str, Any] | None = None,
) -> IResponseList[DataType]:
    payload = IResponseList[DataType](
        data=data,
        meta=meta or {},
    )
    if message is not None:
        payload.message = message
    return payload


# db_engine.py
import asyncio
from typing import Any

from sqlalchemy.ext.asyncio import (
    AsyncEngine,
    AsyncSession,
    async_sessionmaker,
    create_async_engine,
)
from sqlalchemy.pool import NullPool


class TenantDatabaseManager:
    def __init__(self, config: Any):
        self.config = config
        self._engines: dict[str, AsyncEngine] = {}
        self._db_factories: dict[str, async_sessionmaker[AsyncSession]] = {}
        self._locks: dict[str, asyncio.Lock] = {}
        self._locks_guard = asyncio.Lock()

    async def _get_lock(self, tenant: str) -> asyncio.Lock:
        lock = self._locks.get(tenant)
        if lock is not None:
            return lock

        async with self._locks_guard:
            lock = self._locks.get(tenant)
            if lock is None:
                lock = asyncio.Lock()
                self._locks[tenant] = lock
            return lock

    def _build_connect_args(self, tenant: str) -> dict[str, Any]:
        return {
            "options": f"-c application_name={self.config.app_name_prefix}:{tenant}"
        }

    def _create_engine(self, tenant: str) -> AsyncEngine:
        try:
            db_url = self.config.service.get(tenant).db.url.get_secret_value()
            return create_async_engine(
                db_url,
                connect_args=self._build_connect_args(tenant),
                **self._engine_args(),
            )
        except Exception as exc:
            raise RuntimeError(
                f"Failed to create database engine for tenant '{tenant}'"
            ) from exc

    def _engine_args(self) -> dict[str, Any]:
        if self.config.mode == "testing":
            return {
                "echo": self.config.echo_sql,
                "poolclass": NullPool,
            }

        return {
            "echo": self.config.echo_sql,
            "pool_pre_ping": self.config.pool_pre_ping,
            "pool_size": self.config.pool_size,
            "max_overflow": self.config.max_overflow,
            "pool_timeout": self.config.pool_timeout,
        }

    async def get_engine(self, tenant: str) -> AsyncEngine:
        engine = self._engines.get(tenant)
        if engine is not None:
            return engine

        lock = await self._get_lock(tenant)
        async with lock:
            engine = self._engines.get(tenant)
            if engine is None:
                engine = self._create_engine(tenant)
                self._engines[tenant] = engine
            return engine

    async def get_sessionmaker(self, tenant: str) -> async_sessionmaker[AsyncSession]:
        factory = self._db_factories.get(tenant)
        if factory is not None:
            return factory

        lock = await self._get_lock(tenant)
        async with lock:
            factory = self._db_factories.get(tenant)
            if factory is not None:
                return factory

            engine = self._engines.get(tenant)
            if engine is None:
                engine = self._create_engine(tenant)
                self._engines[tenant] = engine

            factory = async_sessionmaker(
                engine,
                class_=AsyncSession,
                expire_on_commit=False,
            )
            self._db_factories[tenant] = factory
            return factory

    async def dispose_tenant(self, tenant: str) -> None:
        engine = self._engines.pop(tenant, None)
        self._db_factories.pop(tenant, None)
        self._locks.pop(tenant, None)

        if engine is not None:
            await engine.dispose()

    async def dispose_all(self) -> None:
        engines = list(self._engines.values())

        self._engines.clear()
        self._db_factories.clear()
        self._locks.clear()

        for engine in engines:
            await engine.dispose()


# from .config import MultitenantConfig
# from .models import TenantConfig, OidcBffConfig, OidcSharedSettings
from .tenant_context import TenantContextConfig
from .tenant_db import TenantDbConfig
from .tenant_runtime import TenantRuntimeConfig
from .db_engine import TenantDatabaseManager
from .models import TenantConfig, OidcBffConfig, OidcSharedSettings, KeycloakAdminConfig
from .resolvers import resolve_oidc_config
from .response_schema import (
    Page, 
    PageBase, 
    PageParams, 
    IResponseBase, 
    IResponseList, 
    IPostResponseBase, 
    IPatchResponseBase, 
    IDeleteResponseBase, 
    IGetResponseBase,
    IGetResponsePaginated,
    IPutResponseBase,
    create_response,
    create_list_response
)

__all__ = [
    "TenantContextConfig",
    "TenantDbConfig",
    "TenantRuntimeConfig",
    "TenantDatabaseManager",
    "TenantConfig", 
    "OidcBffConfig", 
    "OidcSharedSettings",
    "KeycloakAdminConfig",
    "resolve_oidc_config",
    "Page", 
    "PageBase", 
    "PageParams", 
    "IResponseBase", 
    "IResponseList", 
    "IPostResponseBase", 
    "IPatchResponseBase", 
    "IDeleteResponseBase", 
    "IGetResponseBase",
    "IGetResponsePaginated",
    "IPutResponseBase",
    "create_response",
    "create_list_response"
]




# https://chatgpt.com/share/e/69ca1d35-3154-800d-9fa4-c1aed736d70a

from __future__ import annotations

import logging
from datetime import datetime
from typing import Any, Generic, TypeVar, TYPE_CHECKING

from pydantic import BaseModel
from sqlalchemy import Select, delete as sa_delete, exists, func, inspect, select
from sqlalchemy.exc import IntegrityError
from sqlalchemy.ext.asyncio import AsyncSession

from fastapi_pagination import Params
from fastapi_pagination.ext.sqlalchemy import paginate

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from gisweb_tenants.tenant_db import TenantDbConfig


ModelType = TypeVar("ModelType")
CreateSchemaType = TypeVar("CreateSchemaType", bound=BaseModel)
UpdateSchemaType = TypeVar("UpdateSchemaType", bound=BaseModel)
ResultType = TypeVar("ResultType")


class MissingDbError(RuntimeError):
    pass


class CRUDBase(Generic[ModelType, CreateSchemaType, UpdateSchemaType]):
    def __init__(
        self,
        model: type[ModelType],
        *,
        config: TenantDbConfig | None = None,
    ):
        self.model = model
        self.config = config

    # -------------------------
    # DB
    # -------------------------

    def _db(self, db: AsyncSession | None = None) -> AsyncSession:
        if db is not None:
            return db

        if self.config is None:
            raise MissingDbError(f"No active DB for {self.model.__name__}")

        return self.config.current_db()

    # -------------------------
    # AUDIT
    # -------------------------

    def _current_user_id(self) -> str | None:
        if self.config is None:
            return None
        return self.config.current_user_id()

    def _apply_create_audit(self, obj: ModelType) -> None:
        user_id = self._current_user_id()
        if not user_id:
            return

        if hasattr(obj, "created_by") and not getattr(obj, "created_by", None):
            setattr(obj, "created_by", user_id)
        if hasattr(obj, "created_by_id") and not getattr(obj, "created_by_id", None):
            setattr(obj, "created_by_id", user_id)

        if hasattr(obj, "updated_by"):
            setattr(obj, "updated_by", user_id)
        if hasattr(obj, "updated_by_id"):
            setattr(obj, "updated_by_id", user_id)

    def _apply_update_audit(self, obj: ModelType) -> None:
        user_id = self._current_user_id()
        if not user_id:
            return

        if hasattr(obj, "updated_by"):
            setattr(obj, "updated_by", user_id)
        if hasattr(obj, "updated_by_id"):
            setattr(obj, "updated_by_id", user_id)

        if hasattr(obj, "updated_at"):
            setattr(obj, "updated_at", datetime.now())

    # -------------------------
    # HOOKS
    # -------------------------

    def _prepare_create_data(
        self,
        obj_in: CreateSchemaType | dict[str, Any],
    ) -> dict[str, Any]:
        return self._to_dict(obj_in)

    def _prepare_update_data(
        self,
        obj_in: UpdateSchemaType | dict[str, Any],
    ) -> dict[str, Any]:
        return self._to_dict(obj_in)

    async def _before_create(
        self,
        *,
        db_obj: ModelType,
        db: AsyncSession,
    ) -> None:
        return None

    async def _after_create(
        self,
        *,
        db_obj: ModelType,
        db: AsyncSession,
    ) -> None:
        return None

    async def _before_update(
        self,
        *,
        db_obj: ModelType,
        update_data: dict[str, Any],
        db: AsyncSession,
    ) -> None:
        return None

    async def _after_update(
        self,
        *,
        db_obj: ModelType,
        db: AsyncSession,
    ) -> None:
        return None

    async def _before_delete(
        self,
        *,
        db_obj: ModelType,
        db: AsyncSession,
    ) -> None:
        return None

    async def _after_delete(
        self,
        *,
        db_obj: ModelType,
        db: AsyncSession,
    ) -> None:
        return None

    # -------------------------
    # DATA
    # -------------------------

    def _to_dict(self, obj_in: BaseModel | dict[str, Any]) -> dict[str, Any]:
        if isinstance(obj_in, BaseModel):
            return obj_in.model_dump(exclude_unset=True)
        return dict(obj_in)

    # -------------------------
    # PK / COLUMNS
    # -------------------------

    @property
    def pk_column(self):
        mapper = inspect(self.model)
        pk = mapper.primary_key

        if len(pk) != 1:
            raise ValueError(f"{self.model.__name__}: only single-column PK supported")

        return pk[0]

    def get_column(self, name: str):
        columns = self.model.__table__.columns
        if name not in columns:
            raise ValueError(f"{self.model.__name__}: invalid sort column '{name}'")
        return columns[name]

    # -------------------------
    # QUERY HELPERS
    # -------------------------

    def base_select(self) -> Select:
        return select(self.model)

    def apply_ordering(
        self,
        stmt: Select,
        *,
        sort: str | None = None,
        order: str = "asc",
        default_sort: str = "id",
    ) -> Select:
        sort_name = sort or default_sort
        column = self.get_column(sort_name)
        return stmt.order_by(column.asc() if order == "asc" else column.desc())

    def apply_offset_limit(
        self,
        stmt: Select,
        *,
        offset: int = 0,
        limit: int | None = None,
    ) -> Select:
        stmt = stmt.offset(offset)
        if limit is not None:
            stmt = stmt.limit(limit)
        return stmt

    async def count_query(
        self,
        stmt: Select,
        *,
        db: AsyncSession | None = None,
    ) -> int:
        db = self._db(db)
        count_stmt = select(func.count()).select_from(stmt.order_by(None).subquery())
        result = await db.execute(count_stmt)
        return int(result.scalar_one())

    # -------------------------
    # EXECUTION HELPERS
    # -------------------------

    async def execute_scalars(
        self,
        *,
        stmt: Select,
        db: AsyncSession | None = None,
    ) -> list[Any]:
        db = self._db(db)
        result = await db.execute(stmt)
        return list(result.scalars().all())

    async def execute_rows(
        self,
        *,
        stmt: Select,
        db: AsyncSession | None = None,
    ) -> list[Any]:
        db = self._db(db)
        result = await db.execute(stmt)
        return list(result.all())

    async def execute_mappings(
        self,
        *,
        stmt: Select,
        db: AsyncSession | None = None,
    ) -> list[dict[str, Any]]:
        db = self._db(db)
        result = await db.execute(stmt)
        return [dict(row) for row in result.mappings().all()]

    # -------------------------
    # GET
    # -------------------------

    async def get(
        self,
        id: Any,
        *,
        db: AsyncSession | None = None,
    ) -> ModelType | None:
        db = self._db(db)
        return await db.get(self.model, id)

    async def get_required(
        self,
        id: Any,
        *,
        db: AsyncSession | None = None,
    ) -> ModelType:
        obj = await self.get(id, db=db)
        if obj is None:
            raise ValueError(f"{self.model.__name__} with id={id} not found")
        return obj

    async def get_by_ids(
        self,
        *,
        ids: list[Any],
        db: AsyncSession | None = None,
    ) -> list[ModelType]:
        db = self._db(db)
        stmt = self.base_select().where(self.pk_column.in_(ids))
        result = await db.execute(stmt)
        return list(result.scalars().all())

    async def get_multi(
        self,
        *,
        offset: int = 0,
        limit: int = 100,
        stmt: Select | None = None,
        db: AsyncSession | None = None,
    ) -> list[ModelType]:
        db = self._db(db)

        if stmt is None:
            stmt = self.base_select()
            stmt = self.apply_ordering(stmt, default_sort=self.pk_column.name)
            stmt = self.apply_offset_limit(stmt, offset=offset, limit=limit)

        result = await db.execute(stmt)
        return list(result.scalars().all())

    async def get_multi_ordered(
        self,
        *,
        offset: int = 0,
        limit: int = 100,
        sort: str | None = None,
        order: str = "asc",
        stmt: Select | None = None,
        db: AsyncSession | None = None,
    ) -> list[ModelType]:
        db = self._db(db)

        if stmt is None:
            stmt = self.base_select()

        stmt = self.apply_ordering(
            stmt,
            sort=sort,
            order=order,
            default_sort=self.pk_column.name,
        )
        stmt = self.apply_offset_limit(stmt, offset=offset, limit=limit)

        result = await db.execute(stmt)
        return list(result.scalars().all())

    async def get_multi_paginated(
        self,
        *,
        params: Params | None = None,
        stmt: Select | None = None,
        db: AsyncSession | None = None,
    ):
        db = self._db(db)

        if stmt is None:
            stmt = self.base_select()
            stmt = self.apply_ordering(stmt, default_sort=self.pk_column.name)

        return await paginate(db, stmt, params or Params())

    async def get_multi_paginated_ordered(
        self,
        *,
        params: Params | None = None,
        sort: str | None = None,
        order: str = "asc",
        stmt: Select | None = None,
        db: AsyncSession | None = None,
    ):
        db = self._db(db)

        if stmt is None:
            stmt = self.base_select()

        stmt = self.apply_ordering(
            stmt,
            sort=sort,
            order=order,
            default_sort=self.pk_column.name,
        )

        return await paginate(db, stmt, params or Params())

    async def count(
        self,
        *,
        stmt: Select | None = None,
        db: AsyncSession | None = None,
    ) -> int:
        db = self._db(db)
        if stmt is None:
            stmt = self.base_select()
        return await self.count_query(stmt, db=db)

    async def exists(
        self,
        id: Any,
        *,
        db: AsyncSession | None = None,
    ) -> bool:
        db = self._db(db)
        stmt = select(exists().where(self.pk_column == id))
        result = await db.execute(stmt)
        return bool(result.scalar())

    # -------------------------
    # CREATE
    # -------------------------

    async def create(
        self,
        *,
        obj_in: CreateSchemaType | dict[str, Any] | ModelType,
        db: AsyncSession | None = None,
        commit: bool = True,
        refresh: bool = True,
    ) -> ModelType:
        db = self._db(db)

        if isinstance(obj_in, self.model):
            db_obj = obj_in
        else:
            data = self._prepare_create_data(obj_in)
            db_obj = self.model(**data)

        self._apply_create_audit(db_obj)
        await self._before_create(db_obj=db_obj, db=db)

        db.add(db_obj)

        try:
            if commit:
                await db.commit()
                if refresh:
                    await db.refresh(db_obj)
            else:
                await db.flush()
        except IntegrityError:
            await db.rollback()
            raise

        await self._after_create(db_obj=db_obj, db=db)
        return db_obj

    # -------------------------
    # UPDATE
    # -------------------------

    async def update(
        self,
        *,
        db_obj: ModelType,
        obj_in: UpdateSchemaType | dict[str, Any],
        db: AsyncSession | None = None,
        commit: bool = True,
        refresh: bool = True,
        exclude_fields: set[str] | None = None,
    ) -> ModelType:
        db = self._db(db)
        update_data = self._prepare_update_data(obj_in)
        exclude_fields = exclude_fields or set()

        for key, value in update_data.items():
            if key in exclude_fields:
                continue
            if hasattr(db_obj, key):
                setattr(db_obj, key, value)

        self._apply_update_audit(db_obj)
        await self._before_update(db_obj=db_obj, update_data=update_data, db=db)

        db.add(db_obj)

        try:
            if commit:
                await db.commit()
                if refresh:
                    await db.refresh(db_obj)
            else:
                await db.flush()
        except IntegrityError:
            await db.rollback()
            raise

        await self._after_update(db_obj=db_obj, db=db)
        return db_obj

    # -------------------------
    # DELETE
    # -------------------------

    async def delete(
        self,
        *,
        id: Any,
        db: AsyncSession | None = None,
        commit: bool = True,
    ) -> ModelType | None:
        db = self._db(db)

        db_obj = await self.get(id, db=db)
        if db_obj is None:
            return None

        await self._before_delete(db_obj=db_obj, db=db)
        await db.delete(db_obj)

        if commit:
            await db.commit()
        else:
            await db.flush()

        await self._after_delete(db_obj=db_obj, db=db)
        return db_obj

    async def delete_where(
        self,
        *,
        where,
        db: AsyncSession | None = None,
        commit: bool = True,
    ) -> int:
        db = self._db(db)

        stmt = sa_delete(self.model).where(where)
        result = await db.execute(stmt)

        if commit:
            await db.commit()
        else:
            await db.flush()

        return int(result.rowcount or 0)

    # -------------------------
    # SOFT DELETE
    # -------------------------

    async def soft_delete(
        self,
        *,
        id: Any,
        db: AsyncSession | None = None,
        commit: bool = True,
        active_field: str = "is_active",
        deleted_at_field: str = "deleted_at",
    ) -> ModelType | None:
        db = self._db(db)

        db_obj = await self.get(id, db=db)
        if db_obj is None:
            return None

        if hasattr(db_obj, active_field):
            setattr(db_obj, active_field, False)

        if hasattr(db_obj, deleted_at_field):
            setattr(db_obj, deleted_at_field, datetime.now())

        self._apply_update_audit(db_obj)

        db.add(db_obj)

        if commit:
            await db.commit()
            await db.refresh(db_obj)
        else:
            await db.flush()

        return db_obj